/*
 * Chaika2chApi.js - 2ch API for chaika コアモジュール
 *
 * このファイルのあるべき位置：modules/Chaika2chApi.js
 *
 * Copyright (C) 2015
 * 使用条件・ライセンス等については chaika 本体に準じます。
 *
 * 註：このファイルは、オリジナルの bbs2chreader/chaika の構成物ではありません。
 *     この 2ch API for chaika は、オリジナルの bbs2chreader/chaika の
 *     作成者・開発者とは全く別の人物が開発しているものです。
 *     license.txtやその他のソースコード・付属文書などで明記されている
 *     オリジナルの bbs2chreader/chaika の作成者・開発者・寄付者/貢献者などは、
 *     この 2ch API for chaika の開発には一切関与しておりません。
 *
 * Last Modified : 2015-04-14 01:15:20
 */


EXPORTED_SYMBOLS = ["Chaika2chApi"];
Components.utils.import("resource://chaika-modules/ChaikaCore.js");


const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;

const AUTO_AUTH_OFF     = 0;        // sessionIDの自動更新をしない(起動時のみ更新)
const AUTO_AUTH_ACCESS  = 1;        // datAPIサーバへのアクセス直後に自動更新する
const AUTO_AUTH_TIMER   = 2;        // タイマーを使って一定間隔に自動更新する
const AUTH_INTERVAL_MIN = 180;      // 認証間隔の最小値(秒)
const AUTH_INTERVAL_MAX = 86400;    // 認証間隔の最大値(秒)
const RETRY_COUNT_MAX   = 100;      // リトライ回数の最大値
const SESSIONID_LIFETIME = 86400;   // sessionIDの有効期間(秒)
const ASYNC_DELAY = 0.5;


/** @ignore */
function makeException(aResult){
    var stack = Components.stack.caller.caller;
    return new Components.Exception("exception", aResult, stack);
}


/**
 * 2ch API オブジェクト
 * @class
 */
var Chaika2chApi = {

    VERSION: "0.06",


    /**
     * API認証状態
     */
    authorized: false,
    sessionID: "",
    lastAuthTime: 0,
    lastAuthTried: 0,
    authError: "",
    errorCount: 0,
    dontRetry: false,


    /**
     * 設定情報
     */
    pref: null,


    /**
     * sessionID自動更新用タイマー
     */
    _timer: null,


    /**
     * ブラウザ起動時のプロファイル読み込み後に一度だけ実行され、初期化処理を行う。
     * @private
     */
    _startup: function Chaika2chApi__startup(){
        this._timer = new Chaika2chApiTimer(this._authStart.bind(this));
        this.pref = new Chaika2chApiPref(this._onPrefChange.bind(this));
        this.pref.load();
        this._startupAuth();
    },


    /**
     * ブラウザ終了時に一度だけ実行され、終了処理を行う。
     * @private
     */
    _quit: function Chaika2chApi__quit(){
        this.pref.unload();
        this.pref = null;
        this._timer = null;
    },


    /**
     * this.pref の設定に変化があったときに呼ばれるコールバック
     * @param {String} aName 変化があった設定項目名
     * @param {String} aStatus load:初期化 unload:終了処理 change:動作中の変更通知
     */
    _onPrefChange: function Chaika2chApi__onPrefChange(aName, aStatus){
        switch(aName){
            case "2chapi.enabled":
                this._overrideGetHttpChannel(this.pref.enabled);
                // 設定の読み込み順序は不定なので、初期化のときは
                // この時点での_startupAuth()はできない
                if(aStatus != "load"){
                    if(this.pref.enabled){
                        this._startupAuth();
                    }else{
                        this.clearAuth(false);
                    }
                }
                break;
            case "2chapi.auth_url":
            case "2chapi.appkey":
            case "2chapi.hmkey":
            case "2chapi.ct_value":
            case "2chapi.auth_ua":
            case "2chapi.2ch_ua":
                // 認証に関係する設定が変更された場合。
                // ct_value auth_ua 2ch_ua に関しては、
                // 認証への影響は現在確認されていないが一応
                if(aStatus == "change") this.clearAuth(true);
                break;
            case "2chapi.auto_auth":
            case "2chapi.auth_interval":
            case "2chapi.retry_interval":
            case "2chapi.retry_count":
                if(aStatus == "change") this._setAuthTimer();
                break;
            case "2chapi.wake_delay":
                this._timer.wakeDelay = this.pref.wakeDelay;
                break;
        }
    },


    clearAuth: function Chaika2chApi_clearAuth(aAll){
        this.authorized = false;
        this.sessionID = "";
        this.lastAuthTime = 0;
        this.lastAuthTried = 0;
        this.authError = "";
        this.errorCount = 0;
        this.dontRetry = false;

        if(aAll){
            ChaikaCore.pref.setChar("2chapi.session_id", "");
            ChaikaCore.pref.setInt("2chapi.last_auth_time", 0);
            ChaikaCore.logger.debug("Auth: CLEAR_ALL");
        }else{
            ChaikaCore.logger.debug("Auth: CLEAR");
        }
        var os = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        os.notifyObservers(null, "Chaika2chApi:Auth", "CLEAR");

        this._timer.setTimeout(0);  // stop
    },


    auth: function Chaika2chApi_auth(aForceUpdate){
        if(!this.pref.enabled) return;
        if(aForceUpdate){
            this.clearAuth(true);
        }else{
          //if(this._loadAuth()) return;
            this._loadAuth();
        }
        this._authStart();
    },


    _startupAuth: function Chaika2chApi__startupAuth(){
        if(!this.pref.enabled) return;
        if(this._loadAuth()) return;
        this._timer.setTimeout(ASYNC_DELAY);
    },


    _loadAuth: function Chaika2chApi__loadAuth(){
        if(!(this.pref.appKey && this.pref.hmKey &&
             this.pref.authURL && this.pref.apiURL)) return false;

        var sessionID = ChaikaCore.pref.getChar("2chapi.session_id");
        var lastAuthTime = ChaikaCore.pref.getInt("2chapi.last_auth_time");
        if(!sessionID) return false;    // 取得済のsessionIDなし

        // 前回のsessionIDが存在する場合は、常にそれを読み込んでおいた上で
        // 今すぐ更新する必要があるか判断する（_apiAuthNG のコメントも参照）

        this.authorized = true;
        this.sessionID = sessionID;
        this.lastAuthTime = lastAuthTime;
        this.lastAuthTried = lastAuthTime;
        this.authError = "";
        this.errorCount = 0;
        this.dontRetry = false;

        ChaikaCore.logger.debug("Auth: LOAD; authTime:", new Date(lastAuthTime * 1000));
        var os = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        os.notifyObservers(null, "Chaika2chApi:Auth", "LOAD");

        var elapsedTime = Math.round(Date.now() / 1000) - lastAuthTime;
        var interval = (this.pref.autoAuth != AUTO_AUTH_OFF &&
                        this.pref.authInterval) || AUTH_INTERVAL_MIN;

        // 自動更新OFFの場合でも、起動時認証が短時間に連続するのを避けるため、
        // 過去 AUTH_INTERVAL_MIN 秒間に取得したsessionIDがあればそれを再利用する

        // リアルタイムクロックをユーザーが戻すこともありえる
        if(elapsedTime < 0 || elapsedTime >= interval) return false;  // sessionIDの更新必要

        this._setAuthTimer();

        return true;    // sessionIDはフレッシュ
    },


    _authStart: function Chaika2chApi__authStart(){
        if(!(this.pref.appKey && this.pref.hmKey && this.pref.authURL && this.pref.apiURL)){
            ChaikaCore.logger.debug("Auth: STOP");
            var os = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
            os.notifyObservers(null, "Chaika2chApi:Auth", "STOP");
            this._timer.setTimeout(0);  // stop
            return;
        }
        var ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        var authURL = ioService.newURI(this.pref.authURL, null, null).QueryInterface(Ci.nsIURL);

        var httpChannel = ChaikaCore.getHttpChannel(authURL, "auth");
        httpChannel.setRequestHeader("X-2ch-UA", this.pref.x2chUA, false);
        httpChannel.setRequestHeader("Content-Type", "application/x-www-form-urlencoded", false);
        httpChannel = httpChannel.QueryInterface(Ci.nsIUploadChannel);

        var wk, CT = this.pref.ctValue || "";        // failsafe for null
        CT = (wk = CT.match(/^\d{1,10}/)) ? wk[0] :                             // 固定値
             (CT.search(/Random/i) != -1) ? Math.floor(Math.random() * 1e10) :  // ランダム値
                                            Math.round(Date.now() / 1000);      // UNIX時刻
        CT = ("000000000" + CT).slice(-10);     // 10桁に満たない場合は先頭を0で埋める
        ChaikaCore.logger.debug("Auth CT:", CT);

        var HB = this._getHMAC(this.pref.hmKey, this.pref.appKey + CT);

        var strStream = Cc["@mozilla.org/io/string-input-stream;1"]
                .createInstance(Ci.nsIStringInputStream);
        var postString = String("ID=&PW=&KY=" + encodeURIComponent(this.pref.appKey) +
                                "&CT=" + CT + "&HB=" + HB);
        strStream.setData(postString, postString.length);
        httpChannel.setUploadStream(strStream, "application/x-www-form-urlencoded", -1);
        httpChannel.requestMethod = "POST";

        var authListener2chApi = {
            onStartRequest: function authListener2chApi_onStartRequest(aRequest, aContext){
                this._binaryStream = Cc["@mozilla.org/binaryinputstream;1"]
                        .createInstance(Ci.nsIBinaryInputStream);
                this._data = [];
            },
            onDataAvailable: function authListener2chApi_onDataAvailable(aRequest, aContext,
                                                aInputStream, aOffset, aCount){
                this._binaryStream.setInputStream(aInputStream);
                this._data.push(this._binaryStream.readBytes(aCount));
            },
            onStopRequest: function authListener2chApi_onStopRequest(aRequest, aContext, aStatus){
                var data = this._data.join("");
                ChaikaCore.logger.debug("Auth Response:", data);
                if(data.search(/^SESSION-ID=Monazilla\/.+?:/i) == -1){
                    Chaika2chApi._apiAuthNG(data, aRequest);
                    return;
                }
                // セッションIDの本体は、戻ってきた値の「Monazilla/1.00:」以降の部分
                data = data.replace(/^SESSION-ID=Monazilla\/.+?:/i, '').replace(/[\n\r]/g, '');
                Chaika2chApi._apiAuthOK(data);
            }
        };

        httpChannel.asyncOpen(authListener2chApi, null);
        ChaikaCore.logger.debug("Auth: START");
    },


    _apiAuthOK: function Chaika2chApi__apiAuthOK(aSessionID){
        this.authorized = true;
        this.sessionID = aSessionID;
        this.lastAuthTime = this.lastAuthTried = Math.round(Date.now() / 1000);
        this.authError = "";
        this.errorCount = 0;
        this.dontRetry = false;

        ChaikaCore.pref.setChar("2chapi.session_id", this.sessionID);
        ChaikaCore.pref.setInt("2chapi.last_auth_time", this.lastAuthTime);

        ChaikaCore.logger.debug("Auth: OK; Time:", new Date(this.lastAuthTime * 1000));
        var os = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        os.notifyObservers(null, "Chaika2chApi:Auth", "OK");

        this._setAuthTimer();
    },


    _apiAuthNG: function Chaika2chApi__apiAuthNG(aData, aRequest){
        // 失敗した場合は前回のsessionIDを消さずにそのままにしておく。
        // 前回のsessionIDにはまだ有効期間が残っている場合が多く、
        // 何らかの理由で更新ができない場合は、前回のを引き続き使う。

        this.lastAuthTried = Math.round(Date.now() / 1000);
        this.dontRetry = false;

        try{
            if(aRequest.requestSucceeded){
                // 失敗の原因が設定にある場合は何度やっても同じなのでリトライしない
                this.dontRetry = aData.toLowerCase() == "ng (appkey incorrect length)";
                this.authError = "Server Response: " + aData.slice(0,80);
            }else{
                this.authError = "HTTP Status: " +
                    aRequest.responseStatus + " " + aRequest.responseStatusText;
            }
        }catch(ex){
            this.authError = "Unable to Connect: " + aRequest.URI.spec;
        }

        ChaikaCore.logger.debug("Auth: NG; Time:", new Date(this.lastAuthTried * 1000));
        var os = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        os.notifyObservers(null, "Chaika2chApi:Auth", "NG");
        try{    // TODO:設定でOFFにできるようにする
            var alertsService = Cc["@mozilla.org/alerts-service;1"].getService(Ci.nsIAlertsService);
            alertsService.showAlertNotification("", "Chaika 2ch API",
                        "Authorization failed\n" + this.authError, false, "", null);
        }catch(ex){}

        this.errorCount += 1;
        ChaikaCore.logger.debug("Auth Error:", this.authError);
        ChaikaCore.logger.debug("Auth Error Count:", this.errorCount);

        this._setAuthTimer();
    },


    _setAuthTimer: function Chaika2chApi__setAuthTimer(){
        if(this.pref.autoAuth == AUTO_AUTH_TIMER){
            var elapsedTime = Math.round(Date.now() / 1000) - this.lastAuthTried;
            var interval = (this.errorCount == 0) ? this.pref.authInterval :
                           (!this.dontRetry && this.errorCount <= this.pref.retryCount) ?
                           this.pref.retryInterval : 0;
            var delay = (interval == 0) ? 0 :
                        (elapsedTime < 0 || elapsedTime >= interval) ? ASYNC_DELAY :
                        interval - elapsedTime;
            this._timer.setTimeout(delay);
        }else{
            this._timer.setTimeout(0);  // stop
        }
    },


    _onApiResponse: function Chaika2chApi__onApiResponse(aApiStatus){
        if(!this.pref.enabled) return;

        if(aApiStatus.text == "api_invalid_id" || aApiStatus.text == "api_empty_id"){
            this._timer.setTimeout(ASYNC_DELAY);

        }else if(this.pref.autoAuth == AUTO_AUTH_ACCESS){
            var elapsedTime = Math.round(Date.now() / 1000) - this.lastAuthTried;
            var interval = (this.errorCount == 0) ? this.pref.authInterval :
                           (!this.dontRetry && this.errorCount <= this.pref.retryCount) ?
                           this.pref.retryInterval : 0;

            if(interval != 0 && (elapsedTime < 0 || elapsedTime >= interval)){
                this._timer.setTimeout(ASYNC_DELAY);
            }
        }
    },


    /**
     * HMAC SHA-256 の計算
     * @param {String} aKey 計算に使うキー
     * @param {String} aData 対象データ
     * @return {String} HMAC SHA-256 16進文字列形式
     */
    _getHMAC: function Chaika2chApi__getHMAC(aKey, aData){
        var keyService = Cc["@mozilla.org/security/keyobjectfactory;1"]
                         .getService(Ci.nsIKeyObjectFactory);
        var keyObj = keyService.keyFromString(Ci.nsIKeyObject.HMAC, aKey);

        var converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                        .createInstance(Ci.nsIScriptableUnicodeConverter);
        converter.charset = "UTF-8";
        var dataArray = converter.convertToByteArray(aData);

        var cryptoHMAC = Cc["@mozilla.org/security/hmac;1"]
                         .createInstance(Ci.nsICryptoHMAC);
        cryptoHMAC.init(cryptoHMAC.SHA256, keyObj);
        cryptoHMAC.update(dataArray, dataArray.length);
        var binHMAC = cryptoHMAC.finish(false);

        var hexHMAC = [];
        for(var i=0; i<binHMAC.length; i++){
            hexHMAC.push(("0" + binHMAC.charCodeAt(i).toString(16)).slice(-2));
        }
        return hexHMAC.join("");
    },


    /**
     * ChaikaCore.getHttpChannel() へのoverride設定および解除を行う
     * @param {Bool} aOverride trueなら設定／falseなら解除
     */
    _overrideGetHttpChannel: (function(){
        var ChaikaCore_getHttpChannel = null;   // 元の関数を保持

        return function Chaika2chApi__overrideGetHttpChannel(aOverride) {
            if(aOverride){
                if(!ChaikaCore_getHttpChannel){     // override設定
                    ChaikaCore_getHttpChannel = ChaikaCore.getHttpChannel;
                    ChaikaCore.getHttpChannel = Chaika2chApi_getHttpChannel;
                }
            }else{
                if(ChaikaCore_getHttpChannel){      // override解除
                    ChaikaCore.getHttpChannel = ChaikaCore_getHttpChannel;
                    ChaikaCore_getHttpChannel = null;
                }
            }
        };

        /**
         * ChaikaCore.getHttpChannel() をoverrideし、2ch API対応化に必要な処理を行う
         * aURLが2chのdatファイルのURLなら、APIを介したdat取得に必要な前処理を付け加える
         * aURLが2ch.netドメインのその他のURLなら、User-Agent名の偽装とcookieの制御のみを行う
         * @param {nsIURI} aURL nsIHttpChannel を作成する URL
         * @param {String} aAPI APIサーバURLの種別(auth or dat)
         * @return {nsIHttpChannel}
         */
        function Chaika2chApi_getHttpChannel(aURL, aAPI){
            if(!(aURL instanceof Ci.nsIURI)){
                throw makeException(Cr.NS_ERROR_INVALID_POINTER);
            }
            var self = Chaika2chApi;    // この関数内での this は ChaikaCore を指すため

            // 通常のdatURLの他に、旧過去ログサーバrokkaのURLも認識する。
            // そうすることで、thread.jsからはrokkaサーバが生きているように見えるため、
            // thread.jsをほとんど変更せずにRonin+datAPI過去ログ取得に対応できる。

            var datID, hostPath = aURL.host + aURL.path;
            if(!((datID = hostPath.match(/^(\w+)(?:\.2ch\.net|\.bbspink\.com)\/(\w+)\/dat\/(\d+)\.dat$/i)
                       || hostPath.match(/^rokka(?:\.2ch\.net|\.bbspink\.com)\/(\w+)\/(\w+)\/(\d+)\/\?sid=Monazilla%2F.+?%3A(\w+)$/i)
                  ) && self.pref.appKey && self.pref.hmKey && self.pref.apiURL)){
                var httpChannel = ChaikaCore_getHttpChannel.call(this, aURL);   // 元の関数
                if(aAPI){
                    httpChannel.setRequestHeader("User-Agent",
                                (aAPI == "auth") ? self.pref.authUA : self.pref.userAgent, false);
                    httpChannel.loadFlags |= Ci.nsIRequest.LOAD_ANONYMOUS;  // cookie を渡さない
                }else if(aURL.host.search(/(?:\.2ch\.net|\.bbspink\.com)$/i) != -1){
                    if(aURL.path.indexOf("/test/bbs.cgi?guid=ON") == 0){
                        httpChannel.setRequestHeader("User-Agent", self.pref.postUA, false);
                    }else{
                        httpChannel.setRequestHeader("User-Agent", self.pref.userAgent, false);
                        httpChannel.loadFlags |= Ci.nsIRequest.LOAD_ANONYMOUS;
                    }
                }
                return httpChannel;
            }

            // sessionID については未設定(未認証)でも通常モードへフォールバックせずに進み、
            // datAPIサーバにエラーを出させる。
            // RoninのsessionIDを使うと過去ログdatも通常datも両方とも取得できるようになる。

            var sessionID = datID[4] || self.sessionID;

            var ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
            var apiURL = ioService.newURI(self.pref.apiURL, null, null).QueryInterface(Ci.nsIURL);
            var datURLSpec = apiURL.resolve(datID.slice(1,4).join("/"));
            var datURL = ioService.newURI(datURLSpec, null, null).QueryInterface(Ci.nsIURL);

            var httpChannel = Chaika2chApi_getHttpChannel.call(this, datURL, "dat");  // callee
            httpChannel.setRequestHeader("Content-Type", "application/x-www-form-urlencoded", false);
            httpChannel = httpChannel.QueryInterface(Ci.nsIUploadChannel);

            var hobo = self._getHMAC(self.pref.hmKey, datURL.path + sessionID + self.pref.appKey);

            // sessionID も about:config から自由に変更することが可能なので念のため
            var strStream = Cc["@mozilla.org/io/string-input-stream;1"]
                    .createInstance(Ci.nsIStringInputStream);
            var postString = String("sid=" + encodeURIComponent(sessionID) + "&hobo=" + hobo +
                                    "&appkey=" + encodeURIComponent(self.pref.appKey));
            strStream.setData(postString, postString.length);
            httpChannel.setUploadStream(strStream, "application/x-www-form-urlencoded", -1);
            httpChannel.requestMethod = "POST";

            return httpChannel;
        }
    })(),


    /**
     * 2ch API datサーバが返すステータスを調べて返す
     * @param {nsIHttpChannel} aHttpChannel
     * @param {String} aResponseText
     * @return {Object} 2ch API datサーバへの通信でないなら null
     */
    getApiStatus: function Chaika2chApi_getApiStatus(aHttpChannel, aResponseText){
        if(!(aHttpChannel instanceof Ci.nsIHttpChannel)){
            throw makeException(Cr.NS_ERROR_INVALID_POINTER);
        }
        if(aHttpChannel.requestMethod != "POST" ||
           aHttpChannel.originalURI.spec.indexOf(this.pref.apiURL) != 0) return null;

        var apiStatus = { text: null, userStatus: null, threadStatus: null, kako: false };

        var httpStatus;
        try{
            httpStatus = aHttpChannel.responseStatus;
        }catch(ex){
            apiStatus.text = "network_error";
            return apiStatus;
        }
        try{
            // User-Status: 0 (sessionID無効) or 1 (sessionID有効) or 2 (RoninのsessionID有効)
            // ヘッダそのものが無いときは null / 数値として解釈できないときは NaN
            apiStatus.userStatus = parseInt(aHttpChannel.getResponseHeader("User-Status"));
        }catch(ex){}
        try{
            // Thread-Status: 0 (dat取得不可) or 1 (現行スレ) or 2 (dat落ち) or 3 (過去ログ)
            // ヘッダそのものが無いときは null / 数値として解釈できないときは NaN
            apiStatus.threadStatus = parseInt(aHttpChannel.getResponseHeader("Thread-Status"));
        }catch(ex){}

        if((apiStatus.threadStatus === 2 || apiStatus.threadStatus === 3) &&
           (httpStatus == 200 || httpStatus == 206 || httpStatus == 304)){
            apiStatus.kako = true;      // 取得対象がdat落ち/過去ログ
        }

        if(httpStatus == 500 && apiStatus.userStatus === 1 && apiStatus.threadStatus === 0){
            // 通常のsessionIDでdat落ちしているdatを取ろうとした場合
            //  → httpStatus:500, ResponseText:'NG', User-Status:1, Thread-Status:0
            // 本当のサーバエラーで500が返ってくる場合もあるので、他のヘッダもチェックする
            // 注: RoninのsessionIDを使ってもdatを取れなかった場合は
            //      → httpStatus:404, ResponseText:'404 Not Found', User-Status:2, Thread-Status:0
            //     この場合はこのままユーザに通知した方が適切なのでここでは敢えて何もしない
            // NOTE: Range:ヘッダ付きのリクエストで、本来は 416 Requested Range Not Satisfiable が
            // 返るべき状況でも、上と全く同じレスポンスが返ってくるようである(2015/04/12確認)。
            apiStatus.text = "dat_down";

        }else if(httpStatus == 401 && apiStatus.userStatus === 0){
            // sessionIDが無効なとき（有効期限が切れている場合など）
            //  → httpStatus:401, ResponseText:'ng (not authorized)', User-Status:0
            // TODO: RoninのsessionIDを区別してエラーメッセーを出すようにする
            apiStatus.text = "api_invalid_id";

        }else if(httpStatus == 200 && aResponseText){
            // ステータスは200だがdatは返されず、
            // レスポンスボディに短いエラーメッセージが返されるケース。
            // エラーメッセージは現状では全て小文字で返されるようだが、
            // 念のため全て小文字にそろえてマッチングする。
            var responseText = aResponseText.toLowerCase();

            if(responseText == "ng (not valid)"){
                // AppKeyもしくはHMKeyが無効・sessionIDに無効文字が混入している
                apiStatus.text = (this.sessionID.search(/[^0-9A-Z]/i) != -1) ?
                                            "api_invalid_id" : "api_invalid_keys";

            }else if(responseText == "ng (incorrect uid length)"){
                // sessionIDの長さが正しくない
                apiStatus.text = this.sessionID ? "api_invalid_id" : "api_empty_id";

            }else if(responseText == "ng (appkey incorrect length)"){
                // AppKeyの長さが正しくない
                apiStatus.text = "api_invalid_keys";

            }else if(responseText.search(/^ng( \([\w\d\s]+\))?$/) != -1){
                // 上のコードでは認識できないAPI関連エラー
                apiStatus.text = "api_unknown_error";
            }
        }

        // 診断メッセージ
        var msg = "API Status: " + apiStatus.text +
            (aResponseText ? ", ResponseText:'" + aResponseText + "'" : "") +
            ", httpStatus:" + httpStatus + ", UserStatus:" + apiStatus.userStatus +
            ", ThreadStatus:" + apiStatus.threadStatus + ", Kako:" + apiStatus.kako +
            ", URL:" + aHttpChannel.URI.spec;
        if(apiStatus.text == "api_unknown_error"){
            ChaikaCore.logger.warning(msg);
        }else{
            ChaikaCore.logger.debug(msg);
        }

        // APIアクセス時のsessionID更新など
        this._onApiResponse(apiStatus);

        return apiStatus;
    }

};


/**
 * 設定の読み込みと監視を行うオブジェクト
 * @constructor
 * @private
 */
function Chaika2chApiPref(aCallback){
    this._callback = aCallback;
    this._init();
}

Chaika2chApiPref.prototype = {

    enabled: false,
    apiURL: null,
    authURL: null,
    appKey: null,
    hmKey: null,
    ctValue: null,
    userAgent: null,
    postUA: null,
    authUA: null,
    x2chUA: null,
    autoAuth: null,
    authInterval: null,
    retryInterval: null,
    retryCount: null,
    wakeDelay: null,

    _callback: null,
    _branch: null,

    CHAIKA_BRANCH: "extensions.chaika.",    // from ChaikaCore._startup()

    /** 設定読み込み用関数テーブル */
    _loadFunc: {
        // ** APIのON/OFF **
        "2chapi.enabled": function(p,n){ return p.enabled = n ? ChaikaCore.pref.getBool(n) : false; },
        // ** API・認証 **
        "2chapi.api_url": function(p,n){ var v; return p.apiURL = n &&
                (v = ChaikaCore.pref.getChar(n), (v && v.slice(-1) != "/") ? v + "/" : v); },
        "2chapi.auth_url": function(p,n){ return p.authURL = n && ChaikaCore.pref.getChar(n); },
        "2chapi.appkey":   function(p,n){ return p.appKey  = n && ChaikaCore.pref.getChar(n); },
        "2chapi.hmkey":    function(p,n){ return p.hmKey   = n && ChaikaCore.pref.getChar(n); },
        "2chapi.ct_value": function(p,n){ return p.ctValue = n && ChaikaCore.pref.getChar(n); },
        // ** User-Agent **
        "2chapi.useragent": function(p,n){ return p.userAgent = n &&
                ChaikaCore.pref.getChar(n).replace(/^@mozilla$/i, p._getMozillaUA); },
        "2chapi.post_ua": function(p,n){ return p.postUA = n &&
                ChaikaCore.pref.getChar(n).replace(/^@mozilla$/i, p._getMozillaUA); },
        "2chapi.auth_ua": function(p,n){ return p.authUA = n &&
                ChaikaCore.pref.getChar(n).replace(/^@mozilla$/i, p._getMozillaUA); },
        "2chapi.2ch_ua":  function(p,n){ return p.x2chUA = n &&
                ChaikaCore.pref.getChar(n).replace(/^@mozilla$/i, p._getMozillaUA); },
        // ** 自動更新 **
        "2chapi.auto_auth": function(p,n){ var v;
            return p.autoAuth = n && (v = ChaikaCore.pref.getInt(n),
                (v == AUTO_AUTH_ACCESS || v == AUTO_AUTH_TIMER) ? v : AUTO_AUTH_OFF); },
    // ---------------------
        // リトライ時の設定についてはさほど重要ではないので、
        // 設定画面ができるまでは暫定的に固定値を割り当てる
        "2chapi.auth_interval": function(p,n){ var v;
            p.retryInterval = n && (v = ChaikaCore.pref.getInt(n),
                (v <= 0) ? 0 : Math.max(AUTH_INTERVAL_MIN, Math.min(v, 1800)));
            return p.authInterval = n && (
                (v <= 0) ? 0 : Math.max(AUTH_INTERVAL_MIN, Math.min(v, AUTH_INTERVAL_MAX))); },
        "2chapi.retry_count": function(p,n){ return p.retryCount = n && RETRY_COUNT_MAX; },
    // ---------------------
    //  "2chapi.auth_interval": function(p,n){ var v;
    //      return p.authInterval = n && (v = ChaikaCore.pref.getInt(n),
    //          (v <= 0) ? 0 : Math.max(AUTH_INTERVAL_MIN, Math.min(v, AUTH_INTERVAL_MAX))); },
    //  "2chapi.retry_interval": function(p,n){ var v;
    //      return p.retryInterval = n && (v = ChaikaCore.pref.getInt(n),
    //          (v <= 0) ? 0 : Math.max(AUTH_INTERVAL_MIN, Math.min(v, AUTH_INTERVAL_MAX))); },
    //  "2chapi.retry_count": function(p,n){ var v; return p.retryCount = n &&
    //          (v = ChaikaCore.pref.getInt(n), Math.min(v, RETRY_COUNT_MAX)); },
    // ---------------------
        "2chapi.wake_delay": function(p,n){ return p.wakeDelay = n && ChaikaCore.pref.getInt(n); },
    },

    /** @private */
    _init: function Chaika2chApiPref__init(){
        var prefService = Cc["@mozilla.org/preferences-service;1"]
                          .getService(Ci.nsIPrefService);
        this._branch = prefService.getBranch(this.CHAIKA_BRANCH);
    },

    _getMozillaUA: function Chaika2chApiPref__getMozillaUA(){
        return Cc["@mozilla.org/network/protocol;1?name=http"]
               .getService(Ci.nsIHttpProtocolHandler).userAgent;
    },

    observe: function Chaika2chApiPref_observe(aSubject, aTopic, aData){
        if(aTopic != "nsPref:changed") return;

        // 上のテーブルにない 2chapi.session_id などの
        // 変化でも通知が来るため、関数の存在チェックが必要
        if(this._loadFunc[aData]){
            var value = this._loadFunc[aData](this, aData);
            if(value !== undefined){
                ChaikaCore.logger.debug("Pref Change:", aData, "=", value);
            }
            if(this._callback) this._callback(aData, "change");
        }
    },

    load: function Chaika2chApiPref_load(){
        this._branch.addObserver("2chapi.", this, false);

        for(var name in this._loadFunc){
            this._loadFunc[name](this, name);
            if(this._callback) this._callback(name, "load");
        }
    },

    unload: function Chaika2chApiPref_unload(){
        this._branch.removeObserver("2chapi.", this);

        for(var name in this._loadFunc){
            this._loadFunc[name](this, null);
            if(this._callback) this._callback(name, "unload");
        }
    },
};


/**
 * ワンショットタイマー(sleep対応)
 * 注：linuxではPCをsleepさせると正常に動作しない可能性あり
 *     https://bugzilla.mozilla.org/show_bug.cgi?id=758848
 * @constructor
 * @private
 */
function Chaika2chApiTimer(aCallback){
    this.callback = aCallback;
}

Chaika2chApiTimer.prototype = {

    callback: null,
    timer: null,
    delay: 0,
    expire: 0,
    _wakeDelay: 10000,  // ミリ秒単位

    set wakeDelay(aDelay){
        if(aDelay != null && aDelay >= 0){
            this._wakeDelay = aDelay * 1000;
        }
    },

    observe: function Chaika2chApiTimer_observe(aSubject, aTopic, aData){
        switch(aTopic){
            case "timer-callback":
                ChaikaCore.logger.debug("Timer: FIRE");
                this.delay = this.expire = 0;
                this.callback();
                return;
            case "sleep_notification":
                ChaikaCore.logger.debug("Timer: SLEEP");
                this.timer.cancel();
                return;
            case "wake_notification":
                ChaikaCore.logger.debug("Timer: WAKE");
                if(!this.expire) return;
                var remain = this.expire - Date.now();
                if(remain <= this._wakeDelay || remain > this.delay){
                    remain = this._wakeDelay;
                }
                this.setTimeout(remain / 1000);
                return;
        }
    },

    /**
     * タイマー設定
     * @param {Number} aDelay タイムアウトまでの時間(秒)
     *                        0 ならタイマーを止めて後始末も行う
     */
    setTimeout: function Chaika2chApiTimer_setTimeout(aDelay){
        if(aDelay <= 0){    // stop
            this.delay = this.expire = 0;
            if(!this.timer) return;

            this.timer.cancel();
            this.timer = null;
            var os = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
            os.removeObserver(this, "sleep_notification");
            os.removeObserver(this, "wake_notification");
            ChaikaCore.logger.debug("Timer: stop");
            return;
        }

        if(!this.timer){
            this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
            var os = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
            os.addObserver(this, "sleep_notification", false);
            os.addObserver(this, "wake_notification", false);
        }

        this.delay = aDelay * 1000;
        this.expire = Date.now() + this.delay;
        ChaikaCore.logger.debug("Timer: delay", aDelay + "s, expire", new Date(this.expire));

        this.timer.init(this, this.delay, Ci.nsITimer.TYPE_ONE_SHOT);
    }
};
